;;;  Q-learning  (F-fuzzy.ss)
;;;  problem of the ship reach the shore
;;;
;;;  state 
;;;;    F-table +Fuzzy learning

(set! random mtrandom)
(mtrandom-seed 2019.11)

(define (fuzzy n x)
  (cond ( (= n -1) (fuzzy-1 x))
        ( (= n -2) (fuzzy-2 x))
        ( (> x n)  (+ n 1 (- x)))
        ( else  (+ 1 x (- n)))
  )
)
(define (fuzzy-2 x)
   (cond ((> x -1) 0)
         ((< x -2) 1)
         ;;;(else x)
         (else (- -1 x))
   )
)
(define (fuzzy-1 x)
   (cond ((> x 0) 0)
         ((< x -2) 0)
         ((< x -1) (+ 2  x))
         (else (- x))
   )
)

;;;;

(define ship-p 10)   ;;;  meter  start from 10
(define ship-v 0)    ;;;;  
;;; action
(define ship-acc 0)    ;;;;  accel of the ship,  No break  max 2
;;; moving law
(define mu 0.2)
(define kr 0.2)
(define (pos-next p v acc) 
     (- p v (* 0.5 mu acc) (* 0.5 v (- kr)))
)

(define (v-next v acc)
       (+ v (* mu acc) (* v (- kr)))
)
;;;    in  any time 
;;;; point is (/ (- 200 time) (max pos 1) ) 
;;;    if danger at pos <=0 point is -100000
;;;
(define (reward p v )
  (if (and (<= -0.1 p) (<= p  1.0) (<=  0.0 v) (<= v 0.4))
           1
           0
      )
)
   
;;;;  para
(define alpha 1.0)   ;;;  Caution!!!
(define alpha 0.8)   ;;;  Caution!!!
(define alpha 0.25)   ;;;  Caution!!!
(define alpha 0.125)   ;;;  Caution!!!
;;(define alpha 0.0625)
(define gamma 0.95) 
;;;  fuzzy F-table
;;;
;;;  F-table pos(20), v(10) ;;; 200 rules
;;;
(define F-table (make-matrix 24  12 0.0))
;;;  select  acc  random with prob 5% , else max Q
;;; 
;;;  grade-prob's
(define preg (make-vector 4))

(define p-scale 2)
(define v-scale 5)

(define (beta p v)
  (let* ((x 0)(y 0)(u 0)(w 0) (pb ()) (s 0) (n (int-part (* p p-scale))) 
            (nv (int-part (* v v-scale))))
     (set! x (fuzzy n (* p p-scale)))
     (set! y (fuzzy (+ n 1) (* p p-scale)))
     (set! u (fuzzy nv (* v v-scale)))
     (set! w (fuzzy (+ nv 1) (* v v-scale))) 
     (set! s (+ (* x u)(* x w)(* y u)(* y w)))
     (push  (list (/ (* x u) s) n nv) pb) 
     (push  (list (/ (* x w) s) n (+ nv 1)) pb) 
     (push  (list (/ (* y u) s) (+ n 1) nv) pb)
     (push  (list (/ (* y w) s) (+ n 1)(+ nv 1)) pb)
     pb
))
    
;;;
(define (F-value  i j )
  (matrix-ref F-table (+ i 2)j)
)
(define (F-value-set!  i j val )
  (if  (< val 0) (set! val 0))
  (matrix-set! F-table (+ i 2)j val)
)

(define (FF-value p v)
  (let* ((pb (beta p v)) (s 0)( x 0))
    (while (pair? pb)
      (set! x (pop pb))
      (inc s (* (car x) (F-value (cadr x)(caddr x))))
    )
   s
))

(define acc-best 0)

(define (maxF p v )
  (let* ((m -100000000) (acc 0) (memo m)(pn 0)(vn 0))
    (do ((a 0 (+ 1 a )) (val m))
        ((= a 3))
        (set! pn (pos-next p v a))
        (set! vn (v-next v a))
        (set! val (FF-value  pn vn) )
        (if (> val m) (begin (set! m val) (set! acc a)))
    )
   (if (= m memo)
     (set! acc-best (random 3))
     (set! acc-best acc)
   )
   m
))

(define (delta-F r p v )
  (let* ((s 0))
     (set! r (+ r (* gamma (maxF p v))))
    (set! r (- r (FF-value p v ))) 
    (* alpha r)
))

(define (int-part x)
  (set! x (floor x))
  (if (< x -2) -2 x)
)
      
;;; 
(define (FL info)
  (let* ((a 0) (pn 0)(vn 0)(d 0)(time 0) (pb ()))
   (set! ship-p 10)
   (set! ship-v (* (random) 2.0))
   (while (> ship-p 0)
      (set! pb (beta ship-p ship-v))
      (set! a (select-action))
      (set! pn  (pos-next ship-p ship-v a))
      (set! vn  (v-next ship-v a))
      (set! d (delta-F (reward ship-p ship-v) ship-p ship-v ))
      (FL-update d ship-p ship-v pb)
      (set! ship-p pn)
      (set! ship-v vn)
      (set! ship-acc a)
      (inc time)
  )
    (set! d (delta-F (reward ship-p ship-v) ship-p ship-v ))
    (FL-update d ship-p ship-v (beta ship-p ship-v) )
))

(define (F-exec)
  (let* ((a 0) (pn 0)(vn 0)(d 0)(time 0)(p0 0)(v0 0)(cont #t))
   (set! ship-p 10)
   (set! ship-v (* (random) 2))
   (set! p0 ship-p)
   (set! v0 ship-v)
   (while (and cont (>= ship-p 0.1) (< time 40))
      (set! a (max-action))
      (set! pn  (pos-next ship-p ship-v a))
      (set! vn  (v-next ship-v a))
      (set! ship-p pn)
      (set! ship-v vn)
      (set! ship-acc a)
      (set! cont (not (= (reward ship-p ship-v) 1)))
      (inc time)
  )
   (not cont)
))

       
(define (FL-update d p  v  pb ) 
  (let* ((i 0)(j 0) (pro 0)(q 0)(memo pb) (s 0))
     (while (pair? pb)
       (set! q (car (pop pb)))
       (inc s (* q q))
     )
    (set! s (/ 1 s)) ;;;fuzzy5
    ;;;(set! s 1 )  ;;;fuzzy4
    (set! pb memo)
     (while (pair? pb)
       (set! q (pop pb))
       (set! pro (pop q))
       (set! i (car q))
       (set! j (cadr q))
       (if (> pro 0)
        ;;(F-value-set! i j  (+ (F-value i j) (* d pro pro s))) ;;;fuzzy5
        (F-value-set! i j  (+ (F-value i j) (* d  pro s))) ;;;fuzzy
       )
    )
))

(define (select-action)
   (if (= (random 4) 0)
       (random 3)
       (begin
          (maxF ship-p ship-v)
          acc-best
       )
    )
)

(define (max-action)
    (maxF ship-p ship-v)
    acc-best
)

(define (lognum total k)
   (floor (expt total (/ k 10)))  ;;;20
)
            
(define (sim)
   (let* ((k 1) (total 100) (nt (lognum total 1)))
  ;;;;   Note 5000 will be Down the ratio!!!
 ;;;(let* ((k 1) (total 2000) (nt (lognum total 1)))
  (do ((i 0 (+ i 1)))
      ((= i total))
      (when (> i nt)
         (inc k)
         (if (> nt 100)
               (F nt)
         )
        (set! nt (lognum total k))
      )
      (FL #f)
  )
  (F total)
))

(define out (open-output-file (tk-string "bugform.txt")))

(define (F nt)
  (let* ((suc 0)(num 1000))
    (do ((i 0 (+ i 1)))
      ((= i num))
      (if (F-exec)  (inc suc))
    )
  ;;(format out "gakusyuuDkaisu  ~s ratio ~s %~%" nt (/ suc num 0.01))
  ;;(format out "gakusyuuDkaisu   ratio ~s %~%"  (/ suc num 0.01))
  ;;(format out "gakusyuuDkaisu   ratio  %~%"  )  ;;;   this is OK
  ;;(format out "gakusyuuDkaisu     ratio ~s   %~%" 0)  ;;; bug
  ;;; (format out "gakusyuuDkaisu ratio ~s   %~%" 0)  ;;; OK
  ;;; (format out "gakusyuuDkaisu  ratio ~s   %~%" 0)  ;;; BUG
  ;;  (format out "gakusyuuDkaisu  ratio ~a   %~%" 0)  ;;; BUG
   ;;           (format #t "gakusyuuDkaisu  ratio ~a" 0)  ;;; BUG
   (define x  (format #f "gakusyuuDkaisu  ratio ~a test ~%" 0) ) ;;; BUG
   (display x )     ;;;  This is BUG!  format has a problem!
   ;;(display x out)
  (flush out)
))

;;;(trace format)
(sim)

(close-port out)
(bye)

