/****   Math   functions 
   f_math.c 
***/
#define  F_MATH_C

#include "lisp.h"
#include <stdio.h>
#include <math.h>

extern mpq_t mpqwork1, mpqwork2;
extern mpz_t mpzwork1, mpzwork2;
extern ptr genmul(),gendiv(),genadd(),gensub();
extern ptr fixdiv();
extern ptr DBL0;
extern int ODDP(ptr);   /* see biglog.c */

#define  getfloat(x)  (FLONUM(x))
extern double get_float();
#ifdef SYSTEM64
#define MSB        0x8000000000000000  //64bit MSB
#define MSBSINT    0x2000000000000000  //62bit MSB
#else
#define MSB     0x80000000  //32bit MSB
#define MSBSINT 0x20000000  //30bit MSB
#endif
#define DBLMIN   1.5e-323  // min double*3

int gen_eq(ptr,ptr);
int int_order(INT,INT);
int double_order(double,double);

/***  cosh,sinh conflict MACOSX system   -->  ss_cosh,  ss_sinh***/

ptr s_float(int narg,ptr env)
    
 {
    /*   expr 1  ,  -->double  */
        return(_float(*(stack-1)));
 }

int gsign(x)
    Reg ptr x;
 {
   /****   sign of x  :  return 1 if x is negative ****/
    if(SINTP(x)) 
           return ((INT)x<0);
    else if(BIGINTP(x))
     {
       return (mpz_sgn(BIGNUM(x))==-1);
     }
    else if(RATIOP(x))
     {
       return (mpq_sgn(RATIONAL(x))==-1);
     }
    else if(DOUBLEP(x))
        return (FLONUM(x)<0.0);
        /** return ( (int)CAR(x)<0);    sign bit is MSB of double!!! ***/
   error(E_notreal,x);
   return 0; /* Never reach */
 }

int zerop(x)
    Reg ptr x;
 {
   if(SINTP(x))
       return x==ZERO;
   else if(DOUBLEP(x)){
      /***   =との整合性  ***/
     return  gen_eq(x,ZERO);
   }
      /*** return (FLONUM(x)==0.0);  ***/
   else if(COMPLEXP(x))
      return (zerop(CAR(x))&&zerop(CDR(x)));
   else if((int)ATTR(x)&NUMBER)
      return NO;
   else
     error(E_notnum,x);
   return 0; /* Never reach */
 }

ptr minus(x)
    Reg ptr x;
 {
   /****   (- x) ****/
   Reg ptr n;
   INT m;
    if(SINTP(x)){ 
       /***  2014.5.10  Bug fix  **/
        m=((INT)x)>>2;
      if(m==SINTMIN)
           return bignum(-m);
       return (ptr)((~((INT) x))+3);
     }
    else if(BIGINTP(x))
     {
       return longneg(x);
     }
    else if(RATIOP(x))
     {
       psh(x);
       mpq_neg(mpqwork1,RATIONAL(x));
       x= mpq2rational(mpqwork1);
       pull();
       return x;
     }
    else if(DOUBLEP(x))
      {
        //2018.3.10
       psh(x);
       x=MAKEDBL(-FLONUM(x));
       pull();
       return x;
      }
    else if(COMPLEXP(x))
      {
        psh(x);
        n=minus(CAR(x));
        psh(n);
        x=minus(CDR(x));
        x=SCONS(n,x);  //2018.3.10
        ATTR(x)=COMPLEX;
        pull();
        pull();
       return x;
      }
   return error(E_notnum,x);
 }

INT check_big(p)
  Reg ptr p;
{
  /****   get 32 bit integer from  BIGINT  ****/
  /****   if exceed 32bit return 0  ***/
  /***   see check16   **/
  Reg INT n;
    n=0;
    psh(p);
   /** if(mpz_fits_slint_p(BIGNUM(p)))     This is 32 bit even if 64 system! **/
   if(mpz_fits_slong_p(BIGNUM(p))){
      n= mpz_get_si(BIGNUM(p));  /***  get signed long int   see GMP **/
    }
    pull();
  return n;
}

int gen_eq(x,y)
    Reg ptr x,y;
 {
    /****  =,  /=    ***/
   Reg int s;
   Reg double v,z,u;
   
    if(x==y) return YES;
    if(SINTP(x))
        {
           if(SINTP(y)) return NO;
           if(DOUBLEP(y))
             {
               v=(double)GETNUM(x);
               /*** return (FLONUM(y)==v); ***/
               return gen_eq(MAKEDBL(v),y);
             }
           if(BIGINTP(y)||RATIOP(y)) return NO;
           if(COMPLEXP(y))
            {
              if(gen_eq(x,CAR(y))) return gen_eq(CDR(y),ZERO);
              return NO;
             }
          error (E_notnum,y);
       }
   if(SINTP(y)) return gen_eq(y,x); /*** needs , because COMPLEXP(y) below **/
   if(COMPLEXP(x))
      {
        if(COMPLEXP(y))
         {
            if(gen_eq(CAR(x),CAR(y))) return gen_eq(CDR(x),CDR(y));
            else
              return NO;
          }
        if(gen_eq(ZERO,CDR(x))) return gen_eq(CAR(x),y);
         return NO;
      }
  if(COMPLEXP(y))
    {
      if(gen_eq(ZERO,CDR(y))) 
         y=CAR(y);
       else
          return NO;
     }
   /***  x,y are  not COMPLEX ***/
   if(DOUBLEP(x))
      {
         /** printf("gen_eq dbl x %f y %f x-y %e\n",FLONUM(x),get_float(y),FLONUM(x)-get_float(y)); **/
        /*** return (FLONUM(x)==get_float(y)); ***/
        v=FLONUM(x);
        z=get_float(y);
        u=fabs(v-z);
        /*  2020.8.25    use DBLMIN
                2.9e-39 (org) is too big!
           **/
        if(u<DBLMIN) return YES;   /**** this is nearly 0.0 in float sense ****/
        return 0;
        /***  
        v=fabs(v)+fabs(z);
        return ((u/v)<2.3e-16);
        ***/
    /*** double の仮数部は、52 bit ***/
        
       /***   return (0.0==(FLONUM(x)-get_float(y)));    this is no good ***/
       }
   if(BIGINTP(x)) 
     {
       if(SINTP(y)||RATIOP(y)) 
       {
        return NO;
        }
       if(DOUBLEP(y))
         {
        return (FLONUM(y)==get_float(x));
         }
       if(BIGINTP(y)) 
         {
           psh(x);
           psh(y);
           s= (mpz_cmp(BIGNUM(x),BIGNUM(y))==0);
           stack-=2;
           return s;
         }
         error (E_notnum,y);
    }
    if(RATIOP(x))
        {
           if(SINTP(y)||BIGINTP(y)) return NO;
           if(DOUBLEP(y))
             {
                  return (FLONUM(y)==get_float(x));
             }
           if(RATIOP(y)) 
             {
               psh(x);
               psh(y);
               s= (mpq_cmp(RATIONAL(x),RATIONAL(y))==0);
               stack-=2;
               return s;
             }
            error (E_notnum,y);
       }
   error (E_notnum,y);
    return 0; /* Never*/
 }

int gen_order(x,y)
    Reg ptr x,y;
    /*   x>y 1, x==y 0,  x<y  -1  */
 {
   Reg INT n,s;
   Reg ptr p;
   
    if(x==y) return 0;
    if(SINTP(x))
       {
         n=GETNUM(x);
         if(SINTP(y)) 
              return int_order(n,GETNUM(y));
         if(DOUBLEP(y))
              return double_order((double)n,FLONUM(y));
         if(BIGINTP(y)) 
          {
            s=check_big(y);
            if(s) return int_order(n,s);
            return (gsign(y))?1:-1;
          }
         if(RATIOP(y))
          {
            psh(y);
            n=mpq_cmp_si(RATIONAL(y),GETNUM(x),1);
            pull();
            return -n;
          }
        error (E_notnum,y);
       }
    if(DOUBLEP(x))
      {
       return double_order(FLONUM(x),get_float(y));
       }
    if(BIGINTP(x)) 
      {
       if(SINTP(y)) 
       {
        n=check_big(x);
        if(n) return int_order(n,GETNUM(y));
         return (gsign(x))?-1:1;
       }
       if(DOUBLEP(y))
       {
         return double_order(get_float(x),FLONUM(y));
       }
      if(BIGINTP(y)) 
       {
           psh(x);
           psh(y);
           s= mpz_cmp(BIGNUM(x),BIGNUM(y));
           stack-=2;
           return s;
       }
         if(RATIOP(y))
          {
            psh(x);
            psh(y);
            mpq_set_z(mpqwork1,BIGNUM(x));
            s= mpq_cmp(mpqwork1,RATIONAL(y));
            stack-=2;
            return s;
          }
       error (E_notnum,y);
    }
    if(RATIOP(x))
        {
           if(SINTP(y)){
               psh(x);
               s= mpq_cmp_si(RATIONAL(x),GETNUM(y),1);
               pull();
               return s;
               }
           else if(BIGINTP(y)) {
               psh(x);
               psh(y);
               mpq_set_z(mpqwork1,BIGNUM(y));
               s= mpq_cmp(RATIONAL(x),mpqwork1);
               stack-=2;
               return s;
            }
           else if(DOUBLEP(y))
             {
               return double_order(get_float(x),FLONUM(y));
             }
           else if(RATIOP(y)) 
             {
               psh(x);
               psh(y);
               s= mpq_cmp(RATIONAL(x),RATIONAL(y));
               stack-=2;
               return s;
             }
       }
    error (E_notreal,x);
    return 0;  /* Never */
 }
int int_order(n,m)
  Reg INT n,m;
{
  if(n==m) return 0;
  if(n>m)  return 1;
  else
    return -1;
}
int double_order(n,m)
  Reg double n,m;
{
  if(n==m) return 0;
  if(n>m)  return 1;
  else
    return -1;
}
ptr g_ne(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_eq(*(stack-2),*(stack-1))==0)?a_t:a_f;
 }

ptr g_max(int narg,ptr env)
    /*  lexpr version */
 {
    Reg ptr x,pmax,*free;
    if(narg==0) return error(E_shortarg,a_nil);
    free=stack-narg;
    pmax=*free++;
    narg--;
   while(narg)
   {
    x=*free++;
    if(gen_order(pmax,x)<0) pmax=x;
    narg--;
   }
    return pmax;
 }

ptr g_min(int narg,ptr env)
    /*  lexpr version */
 {
    Reg ptr *free,x,pmin;
    if(narg==0) return error(E_shortarg,a_nil);
      free=stack-narg;
    pmin=*free++;
    narg--;
   while(narg)
   {
    x=*free++;
    if(gen_order(pmin,x)>0) pmin=x;
    narg--;
   }
    return pmin;
 }

ptr g_add1(int narg,ptr env)
 {
   /*  any int,double */
    /* register ptr x=*(stack-1),p; */
  checknarg(narg-1);
      return genadd(*(stack-1),ONE);
    
 }

ptr g_sub1(int narg,ptr env)
 {
  checknarg(narg-1);
      return gensub(*(stack-1),ONE);
 }

ptr ABS(x)
    ptr x;
 {
      if(REALP(x))
        {
          if(gsign(x)) return minus(x);
          return x;
        }
      if(COMPLEXP(x))
        {
          return MAKEDBL(sqrt(get_float(norm2(x))));
        }
   return error(E_notnum,x);
 }

ptr g_abs(int narg,ptr env)
 {
  checknarg(narg-1);
   return ABS(*(stack-1));
 }

ptr g_zerop(int narg,ptr env)
    
 {
    Reg ptr x=*(stack-1);
  checknarg(narg-1);
    return PRED(zerop(*(stack-1)));
   
 }

ptr g_minusp(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return PRED(gsign(*(stack-1)));
 }

ptr g_plusp(int narg,ptr env)
    
 {
    Reg ptr x=*(stack-1);
  checknarg(narg-1);
    if(zerop(x)||gsign(x)) return a_f;
   return a_t;
 }

ptr g_sign(int narg,ptr env)
    
 {
   Reg ptr x=*(stack-1);
  checknarg(narg-1);
   if(zerop(x)) return ZERO;
   if(gsign(x)) return MINUSONE;
   return ONE;
 }

