/*** C for Linux  P.136
   sockaddr.c
   04.7.15
   sslisp
   make sockaddr_in data in scheme
***/

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#endif

#include <stdio.h>

#include <string.h>
#include <stdlib.h>
#include "lisp.h"

extern struct sockaddr_in * get_sock_addr(ptr lpath);

extern int pstrcpy(char *,ptr);  // str.c   

int get_portnumber(Uchar *);
ptr make_sockaddr();
int set_port(int ,struct sockaddr_in *);
int get_addr(Uchar *,struct in_addr *);

static Uchar portnum[20];

ptr s_netport(int narg,ptr env){
 ptr host,port;
  if(narg==1){
    port=*(stack-1);
    host=ATOMVAL(a_hostname);
  }
 else{   
    checknarg(narg-2);
     host=*(stack-1);
     port=*(stack-2);
 }
  if(SREGP(host)||!STRINGP(host)) error(E_notstr,host);
  return  make_sockaddr(host,port);
}

/*** str is port number or service name  ***/
ptr make_sockaddr(ptr str,ptr portno){
  Uchar *p;
  Uint len,port;
  struct in_addr *aptr;
  struct sockaddr_in *net;

  net=(struct sockaddr_in *)Mymalloc(sizeof(struct sockaddr_in));
  if(SINTP(portno))
       port=GETNUM(portno);
  else{
        //p=get_string(portno,&len);
  pstrcpy((char *)portnum,portno);  // str.c    sure??
       port=get_portnumber(portnum);
   }
  set_port(port,net);  
  /** printf("setport OK\n"); **/
  p=get_string(str,&len);
  aptr=&(net->sin_addr);
  /*** aptr=(struct in_addr *)Mymalloc(sizeof(struct in_addr));   ***/
  len=get_addr(p,aptr);
  /*** printf("getaddr OK\n"); ***/
  if(len!=0){
       Myfree(net);
       error(E_network,str);
  }
  str=CONS(sock_addr,(ptr)net);
  ATTR(str)=NETSOCKET;
  /*** printf("sockaddr OK\n"); ***/
  return str;
}

int get_addr(Uchar *p,struct in_addr *aptr){
  struct hostent *host;
  struct in_addr *bptr;

  if(inet_aton((char *)p,aptr)==0){
     /*** by name  ***/
      host=gethostbyname((char *)p);
      if(host==NULL){
       return -1;/** error  ***/
     }
     bptr=(struct in_addr *)*host->h_addr_list;
     memcpy(aptr,bptr,sizeof(struct in_addr));
  }
  /*** fprintf(stderr,"addr=%s\n",inet_ntoa(*aptr)); ***/
  return 0;
}
/*** port number,  p=portno or service-name as smtp ***/
int get_portnumber(Uchar *p){
  int  portno;
   struct servent *se;
  if((se=getservbyname((char *)p,"tcp"))==NULL){
      /***  not service,  portno  ***/
      portno=atoi((char *)p);
  /*** fprintf(stderr,"port %d\n",portno); ***/
     return portno;   /*** return portnumber ***/
  }
  else{
         return ntohs(se->s_port);
  }
}

/****  portno must be normal expression not byte-order  ***/
int set_port(int portno,struct sockaddr_in *pserver){
  int byte_order;
  byte_order=htons(portno);
 memset((char *)pserver,0,sizeof(struct sockaddr_in)); /*** clear  **/
  pserver->sin_family=AF_INET;
  pserver->sin_port=byte_order;
  /***
  portno=ntohs(pserver->sin_port);   port number 
  fprintf(stderr,"port= %d\n",portno); 
 ****/
  return portno;
}
/***  getinformation of net ***/
char *info_hostip(struct sockaddr_in *net){
  return (char *)inet_ntoa(net->sin_addr);
}

int info_portnumber(struct sockaddr_in *net){
  return ntohs(net->sin_port);
}

ptr s_hostip(int narg,ptr env)
{
 int acc;
 ptr s;
 char *p;
  checknarg(narg-1);
  s=*(stack-1);
  if(SREGP(s)) error(E_notstream,s);
  if(STREAMP(s)){
    acc=(long)CAR(s);
    if(!(acc&SOCKET_PORT)) error(E_network,s);
     p=info_hostip(get_sock_addr(s));
   }
   else if(NETSOCKETP(s))
     p=info_hostip((struct sockaddr_in *)CDR(s));
   else
       error(E_network,s);
   return CMAKESTR(p);
}

ptr s_hostname(int narg,ptr env)
{
 int acc;
 ptr s;
 char *p;
 struct hostent *host;
struct sockaddr_in *net;

  checknarg(narg-1);
  s=*(stack-1);
  if(SREGP(s)) error(E_notstream,s);
  if(STREAMP(s)){
    acc=(long)CAR(s);
    if(!(acc&SOCKET_PORT)) error(E_network,s);
     net=get_sock_addr(s);
   }
   else if(NETSOCKETP(s))
     net=(struct sockaddr_in *)CDR(s);
   else
       error(E_network,s);
   /*** sethostent(YES); ***/
   /** printf("ip %s\n",p);***/
   host=gethostbyaddr((char *)&(net->sin_addr),sizeof(net->sin_addr),AF_INET);
   if(host==NULL) {
       /**  printf("error %d\n",h_errno);**/
        return a_f;
    }
   return CMAKESTR(host->h_name);
}

ptr s_portnumber(int narg,ptr env)
{
 int acc;
 ptr s;
 char *p;
  checknarg(narg-1);
  s=*(stack-1);
  if(SREGP(s)) error(E_notstream,s);
  if(STREAMP(s)){
     acc=(long)CAR(s);
     if(!(acc&SOCKET_PORT)) error(E_network,s);
      acc=info_portnumber(get_sock_addr(s));
   }
  else if(NETSOCKETP(s))
      acc=info_portnumber((struct sockaddr_in *)CDR(s));
  else
       error(E_network,s);
  return MAKESINT(acc);
}

ptr s_socketportp(int narg,ptr env)
{
 int acc;
 ptr s;
 char *p;
  checknarg(narg-1);
  s=*(stack-1);
  if(SREGP(s)||!STREAMP(s)) return a_f;
  acc=(long)CAR(s);
  if(!(acc&SOCKET_PORT)) return a_f;
  if((acc&listen_ac)==listen_ac) return a_nil;
  return a_t;
}

ptr s_netportp(int narg,ptr env)
{
 int acc;
 ptr s;
 char *p;
  checknarg(narg-1);
  s=*(stack-1);
  if(SREGP(s)||!NETSOCKETP(s)) return a_f;
  return a_t;
}


