/***   utf8.c
 2004.11.25 
 2005.1.15
 2005.1.17   kanji_error control kconv
       NO then only errorset 
 **/

#include "lisp.h"
#include "kanji.h"
extern int jis_unicode(unsigned int);
extern int unicode_jis(unsigned int);
/*
 * ASCII , hankaku katakana,
 * JIS X 0201 (1976) <-> unicode 
 *  by S.Sato
 */

unsigned int jis2unicode(unsigned int jis)
{
    if (jis <= 0x007e) {
	    return jis;
	}
     else if ((0x00a1 <= jis) && (jis <= 0x00df)) {
	return 0xff61 + jis - 0x00a1;
    }
    else
      return jis_unicode(jis);
}

unsigned int unicode2jis(unsigned int uni)
{
    if (uni <= 0x007e) {
            return uni;
        }
     else if ((0xff61 <= uni) && (uni <= 0xff9f)) {
        return 0xa1 + uni - 0xff61;
    }
   else
    return unicode_jis(uni);
}

/****  unicode-> utf8  return integer ,  0 returned if error ***/

unsigned INT unicode_utf8(unsigned int x){
  unsigned int y,z,w;
    if(x<=0x7f) return x;
    if(x<=0x7ff){   /***  2 byte ***/
         y=(x&0x3f)+0x80;
         x=(x>>6)+0xc0;  /*** c2~df **/
         return (x<<8)+y;
    }
    if(x<=0xffff){   /***  3 byte ***/
       z=(x&0x3f)+0x80;
       y=((x>>6)&0x3f)+0x80;
       x=((x>>12)&0xf)+0xe0;  /*** e0~ef **/
     return (x<<16)+(y<<8)+z;
     }
    if(x<=0x1fffff){   /***  4 byte  2019.10.2 ***/
       w=(x&0x3f)+0x80;
       z=((x>>6)&0x3f)+0x80;
       y=((x>>12)&0x3f)+0x80;
       x=((x>>18))+0xf0;   /*** f0~f7  ***/
     return (x<<24)+(y<<16)+(z<<8)+w;  /*** negative!**/
    }
/*** 5 or 6 bytes  perhaps Never occur ***/
return 0; 
}

/****  jiscode-> utf8  return integer ,  0 returned if error ***/
unsigned INT jiscode_utf8(unsigned int jis){
  unsigned int x,y,z;
    if(jis==0) return jis;
    x=jis2unicode(jis);
    return unicode_utf8(x);
}


ptr s_uni_utf8(int narg,ptr env)
{
  unsigned INT x;
  ptr p;
  checknarg(narg-1);
  p=*(stack-1);
  x=GETNUM(p);
  x=unicode_utf8(x);
   //  if x is 4byte, since top byte is f0~f7, return fixnum even if SYSTEM32 
  // if((int) x<0) return bignum((long) x);
  return MAKESINT(x);
}

unsigned int utf8_uni(unsigned INT x)
{
  unsigned INT y,z,w;
    if(x<=0x7f) return x;
    if((x<=0xdfbf)&&(x>=0xc280)){
         y=(x&0x3f);
         x=(x>>8)&0x1f;
         return (x<<6)+y; /*** 11 bit **/
    }
    if((x<=0xefbfbf)&&(x>=0xe0a080)){
     z=(x&0x3f);
     x=(x>>8);
     y=(x&0x3f);
     z=z+(y<<6);  /** 12 bit ***/
     x=(x>>8);
     x=(x&0xf);  /*** e0~ef **/
     return (x<<12)+z;
    }
    if((x<=0xf7bfbfbf)&&(x>=0xf0908080)){
     w=(x&0x3f);
     x=(x>>8);
     z=(x&0x3f);
     x=(x>>8);
     y=(x&0x3f);
     x=(x>>8);
     x=(x&0x7);  /***  **/
     return (x<<18)+(y<<12)+(z<<6)+w; /** 21 bit **/
    }
   return 0; /*** not in range ***/
}

ptr s_utf8_uni(int narg,ptr env)
{
  unsigned int x;
  ptr p;
  checknarg(narg-1);
  p=*(stack-1);
  if(SINTP(p))
    x=GETNUM(p);
  else if(BIGINTP(p)){
     x=mpz_get_si(BIGNUM(p));
  }
  else
     error(E_range,p);
    return MAKESINT(utf8_uni(x));
}

ptr s_euc_utf8(int narg,ptr env)
{
  Reg ptr new,out;
  Reg unsigned char *p;
  ptr x;
  int n;
  int c,d,kconv_error=0;
  checknarg(narg-1);
  x=*(stack-1);
  if(SREGP(x)||!STRINGP(x)) error(E_notstr,x);
  p=get_string(x,(Uint *)&n);
  new=makestring(n);
  psh(new);
  out=open_str(new,write_ac);
  *(stack-1)=out;

  while(n--){
    if(n<0) {
        kconv_error=1;
        break;
     }
    c=*p++;
    if(c<0x80) {
        str_writec(out, c);
     }
    else if(euc_first(c)){
        d=*p++;
        n--;
        if(c==0x8e)
             c=d;   /*** half kana  ***/
        else
              c=(c<<8)+d-0x8080;
        c=jiscode_utf8(c);
        if(c==0){
               kconv_error=1;
            if(kanji_error) break;
          str_writec(out,'?');
          str_writec(out,'?');
        }
       d=(c>>16);
       if(d)  str_writec(out, d);
       d=(c>>8);
       if(d)  str_writec(out, d&0xff);
        str_writec(out,c&0xff);
    }
    else{
          kconv_error=1;
         if(kanji_error) break;
          str_writec(out,'?');
      }
  }
  new=close_str(out);
  pull();
 if(kconv_error){
       if(kanji_error) error(E_kconv,new);
       ATOMVAL(a_errno)=MAKESINT(E_kconv);
  }
   return new;
}

ptr s_euc_utf8code(int narg,ptr env)
{
  ptr code;
  unsigned INT x,c,d;
  checknarg(narg-1);
  code=*(stack-1);
  x=GETNUM(code);
    if(x<0x80) {
        return code;
     }
   d=x&0xff;
   c=(x>>8)&0xff;
   if(euc_first(c)){
        if(c==0x8e)
             c=d;   /*** half kana  ***/
        else
              c=(c<<8)+d-0x8080;
        x=jiscode_utf8(c);
        //if((int) x<0) return bignum((long) x);
       return MAKESINT(x);
   }
   return ZERO;
}

ptr s_utf8code_euc(int narg,ptr env)
{
  ptr code;
  unsigned int x,c,d;
  checknarg(narg-1);
  code=*(stack-1);
  x=GETNUM(code);
  if(x<=0x7f) return code;
  x=utf8_uni(x);
  x=unicode2jis(x);
  if(x<=0xff) {  /*** hankaku kana  ***/
     x=0x8e00+x;
     return MAKESINT(x);
  }
     x=0x8080+x;
     return MAKESINT(x);
}  

ptr s_utf8code_jis(int narg,ptr env)
{
  ptr code;
  unsigned int x,c,d;
  checknarg(narg-1);
  code=*(stack-1);
  x=GETNUM(code);
  if(x<=0x7f) return code;
  x=utf8_uni(x);
  x=unicode2jis(x);
     return MAKESINT(x);
}


ptr s_utf8_euc(int narg,ptr env)
{
  Reg ptr new,out;
  Reg unsigned char *p;
  ptr x;
  int n;
  int c,d,e,kconv_error=0;
  checknarg(narg-1);
  x=*(stack-1);
  if(SREGP(x)||!STRINGP(x)) error(E_notstr,x);
  p=get_string(x,(Uint *)&n);
  new=makestring(n);
  psh(new);
  out=open_str(new,write_ac);
  *(stack-1)=out;

  while(n--){
    if(n<0) {
        kconv_error=1;
        break;
     }
    c=*p++;
    if(c<0x80) {
        str_writec(out, c);
     }
    else if(c<0xc0){
        kconv_error=1;
        if(kanji_error) break;
        str_writec(out, '?');
     }
    else if(c<0xe0){
        d=*p++;
        n--;
        c=((c&0x1f)<<6)+(d&0x3f);
        c=unicode2jis(c);
        if(c==0){
              kconv_error=1;
              if(kanji_error) break;
              str_writec(out, '?');
        }
        if(c>256){
          str_writec(out, ((c>>8)&0xff)+0x80);
        }
        str_writec(out, (c&0xff)+0x80);
        }
    else if(c<0xf0){
         d=*p++;
        n--;
         e=*p++;
        n--;
        c=((c&0xf)<<12)+((d&0x3f)<<6)+(e&0x3f);
        c=unicode2jis(c);
        if(c==0){
              kconv_error=1;
              if(kanji_error) break;
              str_writec(out, '?');
              str_writec(out, '?');
        }
        else {
            if(c>256){
              str_writec(out, ((c>>8)&0xff)+0x80);
              }
            str_writec(out, (c&0xff)+0x80);
        }
     }
    else{
          kconv_error=1;
          if(kanji_error) break;
          str_writec(out, '?');
      }
  }
  new=close_str(out);
  pull();
 if(kconv_error){
      if(kanji_error) error(E_kconv,new);
      ATOMVAL(a_errno)=MAKESINT(E_kconv);
  }
   return new;
}

ptr EUC_UTF(ptr p){
 ptr s;
 psh(p);
 s=s_euc_utf8(1,a_nil);
 pull();
 return s;
}

ptr UTF_EUC(ptr p){
 ptr s;
 psh(p);
 s=s_utf8_euc(1,a_nil);
 pull();
 return s;
}

ptr s_jis2unicode(int narg,ptr env)
{
  ptr s;
  INT j;
  checknarg(narg-1);
  s=*(stack-1);
  if(!SINTP(s)) error(E_notnum,s);
  j=GETNUM(s);
  j=jis2unicode(j);
  return MAKESINT(j);
}

ptr s_unicode2jis(int narg,ptr env)
{
  ptr s;
  int j;
  checknarg(narg-1);
  s=*(stack-1);
  if(!SINTP(s)) error(E_notnum,s);
  j=GETNUM(s);
  j=unicode2jis(j);
  return MAKESINT(j);
}

ptr s_jiscode2utf(int narg,ptr env)
{
  ptr s;
  INT j;
  checknarg(narg-1);
  s=*(stack-1);
  if(!SINTP(s)) error(E_notnum,s);
  j=GETNUM(s);
  j=jiscode_utf8(j);
  return MAKESINT(j);
}

