;;;  opt-var.ss
;;;;  variable ->frame(stack)
;;;;  次の変数はframe可能である。
;;;; :BIND ただし、内側BLOCKにあるlambda式から参照がないこと。
;;;; :LAMBDA  ただし、内側BLOCK内のLAMBDA式から参照のないこと
;;;; 従って、最初にlambda式の内部を調べて、外側の:ENV変数の参照
;;;  があれば、それをlistにする必要がある。
;;; これを、
;;;   *access-out*
;;;  とする。
;;;  :LAMBDA 変数の場合、渡された、stackのbaseを:FRAMEとして、
;;;  それで、アクセスする。
;;;;　LAMBDAが、(x y . z)のような場合、
;;;;    x = :FRAME 0 , y= :FRAME 1, z= :FRAME 2,
;;; として、:FRAME 3からはBLOCK用にする。
;;;; BLOCKの変数は、
;;;  をbaseとして計算する。
;;; ただし、
;;;  必要な、frameを確保するために、
;;;  (:ALLOC n)
;;;  (:DEALLOC n)
;;;　これらは、compiled listの先頭と最後、
;;; LAMBDA BLOCKの最初と最後におかれる。
;;; :FRAMEは、call/ccで保存しないといけない。

(in-package "compiler")

(define   *access-out* ())
(define   *frame-table* ())

(define (make-out-access )
   (let ((v *comp-vector*)(n *comp-size*)(x 0)
            (lambda-stack ()) (top 0)(block #f)(var 0) (env ())(env-stack ()))
      (set! *access-out* ())
     (do ((i 0))
         ((<= n i) )
         (set! x (vector-ref v i))
          ;;;(format #t "access ~s~%" x)
         (set! top (car x))
         (cond  ((eq? top :LAMBDA-BEGIN)
                    (push block lambda-stack) 
                    (set! block (list (cadr x)))
                  )
                ((eq? top :LAMBDA-END)
                    (set!  block (pop lambda-stack))
                  )
                ( (not block ))  ;;; No op out of lambda
                ((eq? top :BLOCK-IN)
                   (set! block (cons (cadr x) block));;; block in lambda
                  )
                ((eq? top :BLOCK-OUT)
                    (pop block)
                 )
                ((or (eq? top :LOAD)(eq? top :SET))
                   (when (eq? (cadr x) :ENV)
                      (set! var (caddr x))
                      (if (not (member (cdr var) block) )
                         (push var *access-out*)
                      )
                    )
                   )
               (else )
          )
         (set! i (+ i 1))
      )
))

(define (macro-begin? x)
  (eq? (caddr x) :MACRO)
)

(define (change-frame-var )
   (let ((v *comp-vector*)(n *comp-size*)(x 0) (frame-stack ())
            (frame 0) (lambda-stack ()) (y 0) (frame-max 0)
            (block 0) (top 0)(op 0)(var 0)(vars 0) )
      (set! *frame-table* ())
     (do ((i 0))
         ((<= n i) )
         (set! x (vector-ref v i))
         (set! top (car x))
         (set!  op #f)
         (cond  ((eq? top :LAMBDA-BEGIN)
                   (push (cons frame frame-max) frame-stack)
                   (set! frame 0)
                   (set! frame-max  0)
                   (set! block (cadr x))
                   (push (cons x 0)  lambda-stack)
                  )
                ((eq? top :LAMBDA-BIND)
                    (set! op  (lambda-vars-stack (cadr x) block ))
                    (set! frame (car op))
                   (set! frame-max  frame)
                    (set-cdr! (car lambda-stack) frame)
                    (set! op (cdr op))
                  )
                ((eq? top :LAMBDA-END)
                     (set! y (pop lambda-stack))
                     (set! frame (cdr y))
                     (set! y (car y))
                     (set! frame (list :ALLOC (- frame-max frame)))
                     (set-cdr! (cddr y) frame)
                     (if (macro-begin? y)
                        (set-cdr! (cdr x) (cons :MACRO frame))
                        (set-cdr! (cdr x) frame)
                     )
                     (set! frame (pop frame-stack))
                     (set! frame-max (cdr frame))
                     (set! frame (car frame))
                  )  
                ((or (eq? top :LOAD)(eq? top :SET))
                   (when (eq? (cadr x) :ENV)
                      (set! var (caddr x))
                      (set! op (var-frame var x))
                    )
                  )
               ((eq? top :BLOCK-IN)
                   (push frame frame-stack)
                 )
               ((eq? top :BLOCK-OUT)
                  (set! frame (pop frame-stack))
                 )
               ((eq? top :BIND)
                  (set! op (bind-frame (cadr x) frame block))
                  (when  op 
                      (set! frame (+ frame 1))
                      (if (> frame frame-max) (set! frame-max frame))
                  )
                )
               (else )
          )
          (if op
               (vector-set! v i op)
          )

         (set! i (+ i 1))
      )
    (if (> frame-max 0) 
         (begin
           (vector-set! v (- n 2) (list :DEALLOC frame-max));;; need for compile-file
           (vector-set! v 0 (list :ALLOC frame-max))
         )
           (vector-set! v 0 NOP)
   )
))

(define (lambda-vars-stack vars block)               
  (let ((i 0)(k 0)(v 0)(res ())(orig vars) (simple #t))
    (while (pair? vars)
      (set! v (cons (car vars) block))
      (if (member v *access-out*)
           (push (car v) res)
         (begin
           (push #f res)
           (set! k (+ k 1))
           (push (cons v (list :FRAME (cons i block))) *frame-table*)
         )
      )
      (set! i (+ i 1))
      (set! vars (cdr vars))
    )
    (when (symbol? vars)
       (set! simple #f)
       (set! v (cons vars block))
      (when (not (member v *access-out*))
        (push (cons v (list :FRAME (cons i block))) *frame-table*)
        (set! vars #f)
        (set! k (+ k 1))
      )
       (set! i (+ i 1))
    ) 
    (set! vars (append  (reverse res) vars))
    ;;;(set! vars (if (and (= i k) simple) NOP (list :LAMBDA-BIND vars)) )
    (set! vars  (list :LAMBDA-BIND vars orig)) ;;; need for narg check
    (cons i vars)
))  

(define (var-frame var x)
  (let ((s 0))
    (set! s (assoc var *frame-table*))
    (if s
         (cons (car x) (cdr s))
         #f
    )
))

(define (bind-frame v frame block)
  (let ((s 0))
    (if (member v *access-out*)
         #f
       (begin
           (set! s (list :FRAME (cons frame block)))
           (push (cons v s) *frame-table*)
           (cons :SET  s)
       )
    )
))

;;;;;;    (:LOAD :IMMIDIATE <#undef>)
;;;;;   -> (:LOAD :UNDEF)
;;;;;      (:LOAD :T)         
;;;;;      (:LOAD :F)         
;;;;;      (:LOAD ())
;;;;;      (:LOAD 0)
;;;;;      (:LOAD 1)
;;;;;      (:LOAD -1)
(define (load-imm-change)
   (let ((v *comp-vector*)(n *comp-size*)(x 0) (val 0))
     (do ((i 0))
         ((= n i) )
         (set! x (vector-ref v i))
          (when (and (eq? (car x) :LOAD) (eq? (cadr x) :IMMIDIATE))
               (set! val (caddr x))
              (cond ((eq? val undef) (set! val :UNDEF)) 
                    ((eq? val #t) (set! val :T)) 
                    ((eq? val #f) (set! val :F)) 
                    ((eq? val ()) ) 
                    ((and (integer? val)(<= val 1)(>= val -1))
                       )
                    (else (set! val #f))
               )
              (if val
                 (vector-set! v i (list :LOAD val))
              )
          )
         (set!  i(+ i 1))
     )
 )
)
         
