;;;;  04.1.12
;;;;      separated from pat.ss
;;;; 04.4.04  New macro version
;;;; 05.11.3  New
;;;;  10.7.5  Ver5 start

(in-package "macro5")
#|
(define define-syntax
   (macro (key exp)
         `(define ,key
             (->syntax (begin (set! *macro-key* ',key) ,exp) )
          )
   )
)
|#

(define *exflag* 0)
(define *local-env* ())
;;;;(define *macro-bind* ())

(define (syntax-error p)
   (error 1000 (cons "illegal syntax" p))
)

(define (inc-exflag)
   (set! *exflag* (+ *exflag* 1))
   (if (> *exflag* 1000) (error 5000 "*** macro-expand over limit ****"))
)
;;;  define-syntax   ---> psyntax.ss
;;;;  only for global macro
(define (get-macro  key)
 (let* ((d ()))
    (if (symbol? key)
      (if (symbol-bind?  key) (set! key (symbol-value key))) 
    )
    (if (or (syntax-lambda? key)(classic-macro? key))
     key
     #f
    )
))
(define macro? get-macro)

;;;  change key->new in tree
(define (exchange-tree lis key new)
  (if (pair? lis)
      (cons (exchange-tree (car lis) key new)
            (exchange-tree (cdr lis) key new) )
      (if (eq? lis key) 
              new
              lis
      )
  )
)


(define (syntax-expand f env)
   (let* ((memo *local-env*))
    (set! *exflag* 0)
    (set! *local-env* env)
    (set! f (syntax-ex f ))
    (set! *local-env* memo)
    f 
))

(define f syntax-expand)

(define (syntax-ex f )
  (if (pair? f)
        (syntax-ex-pair f)
       f
  )
)

;;;;  Non call/cc version
(define (syntax-ex-pair  f )
  (let* ( (top (car f))(x 0)(c *exflag*) )
    ;;;(format #t "syntax-ex start ~s ~s~%" f *local-env*)
       (cond ( (pair? top)  ;;; ((lambda  ...))  may exists macro in the form
                (set! x (syntax-ex top ))
                 (if (not (equal? top x))
                     (syntax-ex-pair  (cons x (cdr f) ))
                     (syntax-ex-args f )
                 )
               )
    ;;;(format #t "syntax-ex 1  ~s ~s~%" top *local-env*)
            ( (special? top)
               ;;;(format #t "syntax-ex-pair special  ~s ~%" top )
               ;;;(set! top (syntax-special? top ))  need???
               (set! top (special? top ))
               (special-ex top f))
            ( (symbol? top)
                (syntax-ex-symbol-pair f)  )
            ( (syntax-lambda? top)
               (inc-exflag)
              (syntax-ex (syntax-lambda-ex f top)))
            ((classic-macro? top)
               (inc-exflag)
               (syntax-ex (macro-expand f))
                )
            (else
                  ;;; we assume function call
                  (syntax-ex-args f )
               )
        )
))

#|
;;;;  This is need???
(define (syntax-ex-classic f)
   (let* ((top (car f)) (x 0))
    ;;; may be top is special-syntax
    ;;;  here  top is  special-macro
    ;;; top is not syntax-macro
     (set! x (classic-macro? top))
     (if x
          (begin
              (inc-exflag)
              (syntax-ex (macro-expand f))
          )
          (begin
              (set! top  (syntax-special? top ))
               (if top
                    (special-ex top f )
                    (syntax-ex-args f )
              )
          )
      )
))
|#

(define (syntax-ex-symbol-pair  f )
  (let* ( (top (car f))(x 0)(v 0)(c *exflag*) )
   ;;; (format #t "syntax-ex 2  ~s ~s~%" top *local-env*)
          ;;; top is symbol 
         (set! v  (assq top *local-env*))
       (if v
             (set! x (cdr v))
             (set! x (get-macro top ))  ;;;new
       )
       (if  (syntax-lambda? x)
              (syntax-ex (syntax-lambda-ex f x))
            (if (classic-macro? x)
              (syntax-ex (macro-expand f ))
               (syntax-ex-args  f )
             )
       )
))
      ;;; (syntax-ex-classic f)

(define (syntax-ex-args  f  )
   (let* ( (res ())(x 0)(y 0) (ans #f)(args (cdr f)) )
     (push (car f) res)
     (while (pair? args)
        (set! x (car args))
        (set! y(syntax-ex  x ))
        (push y res)
        (if (not (equal? x y)) (set! ans #t))
        (set! args (cdr args))
     )
    (if ans
      (begin
        (push args res)   ;;;  may be dot-list
        (apply list* (reverse res))
        ;;;;(reverse res)
      )
     f
    )
))

(define  (let*-arg-ex  args  )
  (let* ((x 0) (res ()) (narg ())(init 0) (env *local-env*))
    (while (pair? args)
       (set! x (car args))
       (set! init (cadr x))
       (set! x (car x))
       (set! init (syntax-ex init ))
       (push (cons x init) env)
       (push (list x init) narg)
        (set! args (cdr args))
       (set! *local-env* env)
    )
    ;;;(format #t "let*-arg-ex out ~s local-env ~s ~%" (reverse narg) env)
    (reverse narg) 
))

(define  (let-arg-ex  args  )
  (let* ((x 0) (res ()) (narg ())(init 0) (env *local-env*))
    (while (pair? args)
       (set! x (car args))
       (set! init (cadr x))
       (set! x (car x))
       (set! init (syntax-ex init ))
       (push (cons x init) env)
       (push (list x init) narg)
        (set! args (cdr args))
    )
       (set! *local-env* env)
    ;;;(format #t "let-arg-ex out ~s local-env ~s ~%" (reverse narg) env)
    (reverse narg)
))

(define  (letrec-arg-ex  args  )
  (let*  ((x 0) (res ()) (narg args)(nenv *local-env*)(env nenv)(init 0))
    (while (pair? args)
       (set! x (car args))
       (set! init (cadr x))
       (set! x (car x))
       (push (cons x init) env)
       (set! args (cdr args))
     )
    (set! *local-env* env)
    (set! args narg)
    (set! narg ())
    (while (pair? args)
       (set! x (car args))
       (set! init (cadr x))
       (set! x (car x))
       (set! init (syntax-ex init))
       (push (list x init) narg)
       (push (cons x init) nenv)
       (set! args (cdr args))
    )
    (set! *local-env* nenv)
    (reverse narg) 
))

(define  (lambda-arg-env args )
 (let* ((env *local-env*))
    (while (pair? args)
      (push (cons (car args) #f) env)
      (set! args (cdr args))
    )
    (if (not (null? args))
        (push (cons  args #f) env)
    )
   (set! *local-env* env)
 env
))

(define  (special-ex top f )
  (let* ((args 0)(g 0)(env *local-env*) )
   ;;;(format #t "special-ex ~s f ~s env ~s~%" top f env)
   (cond ( (eq? top 'let*)
            (set! args (let*-arg-ex  (cadr f) ))
  ;;; (format #t "special-ex inside let* ~s f ~s env ~s~%" top f *local-env*)
            (set! g (syntax-ex-args (cdr f)  ))
            (set! f(cons (car f) (cons args (cdr g))))
           ;;; (set! f (name-change-close f ))
            (set! *local-env* env)
           ;;; (set! *name-change* name)
  ;;; (format #t "special-ex out  let*    f ~s ~%"  f )
            f
          )
         ( (eq? top 'letrec)
            ;;;(set! *name-change* ())
            (set! args (letrec-arg-ex  (cadr f) ))
            (set! g (syntax-ex-args (cdr f) ))
            (set! f(cons (car f) (cons args (cdr g))))
           ;;; (set! f (name-change-close f ))
            (set! *local-env* env)
            f
            )
         ( (eq? top 'let)
            (set! args (let-arg-ex  (cadr f) ))
                ;;; (format #t "special-ex inside let* ~s f ~s env ~s~%" top f *local-env*)
            (set! g (syntax-ex-args (cdr f)  ))
            (set! f(cons (car f) (cons args (cdr g))))
           ;;; (set! f (name-change-close f ))
            (set! *local-env* env)
           ;;; (set! *name-change* name)
  ;;; (format #t "special-ex out  let*    f ~s ~%"  f )
            f
          )
         ((eq? top 'define)   ;;;;;mada 問題あり
                              ;;;;  env
                              ;;;; (define (f x) ...)
                              ;;;;  need top-env?
              ;;;; 04.3.15
             (set! g  (cddr f))
             (set! args (cadr f))
             (if (pair? args)
                  (begin
                    (set! g (list* 'lambda (cdr args) g)) 
                    (set! args (car args))
                    (inc-exflag)
                  )
                 (if (null? (cdr g))
                    (set! g (car g)) ;;; (define x g))
                    (error 4443 f)
                 )
             )
             (set! g (syntax-ex g))
             (set! g (list top args g))
             (format #t "special-define:~s~%" g)
             (push (cons args g) *local-env*)   ;;; V2.0
             g
           )
         ((eq? top 'set!) 
             (set! g (syntax-ex (caddr f) ))
             (list (car f) (cadr f) g)
           )

         ((or (eq? top 'quote)(eq? top 'INLINE))  ;;; V2.0
             f
           )
        ((or (eq? top 'begin) (eq? top 'if))
             (syntax-ex-args f  )
            )
        ( (or (eq? top 'lambda) (eq? top 'macro))
            (lambda-arg-env (cadr f) )
            (set! g (syntax-ex-args (cdr f)  ))
            (set! *local-env* env)
            (cons (car f) g)
          )
        (else
           (error 4444 f)
        )
     )
))
#|
(define (ex f )
  (let* ((def ()))
    (set! def (get-macro (car f) ))  ;;; New
    (if def
      (macro-ex f  def)
       f
    )
))
|#
;;;;  
(define (syntax-lambda-ex f def )
  (let* ((pat 0)(env ())(rule 0)(rules 0)(ans #f))
    (inc-exflag)
    ;;;(eval (list (caddr  def) (list quote f)) *local-env*)
    ;;;(eval (list (cdr  def) (list quote f)) *local-env*)
    (set! f (eval (list  (<-syntax  def) (list quote f)) *local-env*))
    (->syntax def)  ;;;; This need. because, eval call syntax-expand
                    ;;;;  and causes infinity loop!
    f
))

;;;; New! 
;;;  *name-change*   assoc list of (var . alias)
#|
(define (name-change-close f )
   (if  (pair? *name-change*)
     (let* ((x 0)(arg ())(v *name-change*) (var 0))
       (while (pair? v)
         (set! x (car v))
           (set! var (car x))
           (set! arg (cons (list (cdr x) var) arg))
         (set! v (cdr v))
      )
        ;;;;(list *system-let** arg f)
        (list (special-alias 'let*)  arg f)  ;;;05.5.7 
     )
     f
   )
)
|#


;;;  need  (interactin-environment)!!!!

(define (syntax-special? x . option)
  (let* ((v 0)(env *local-env*) (y 0))
    (if (not (eq? option ()))  (set! env (car option)))
   ;;; (format #t "syntax-special?  x ~s env ~s~%" x env)
    (set! v (assq x env))
     (if v 
       (set! x (cdr v))
       (begin
        (set! y (special? x))
        (if y (set! x y))
       )
    )
    (if (memq x '(quote let* letrec set! if define lambda macro call/cc begin let-syntax letrec-syntax))
        x
       #f
     )
))

(define (s1 f)
    ( syntax-expand  f ())
)

;;;New!
(define (quote-recover f  data)
  (let* ((s 0) (qqq 0))
    (if (pair? f)
       (begin
          (set! qqq (car f))
          (set! s (special? qqq))
          (if (and s (eq? s 'quote))
              (cons qqq (quote-recover-exec (cdr f) data))
              (cons (quote-recover qqq data)
                    (quote-recover (cdr f) data))
          )
       )
      f
     )
))

(define (quote-recover-exec f data)
  (let* ((s 0))
    ;;; (format #t "f ~s data ~s~%" f data)
    (if (pair? f)
           (cons (quote-recover-exec (car f) data)
                    (quote-recover-exec  (cdr f) data))
         (begin
           (set! s (special? f))
           (if s
               s
             (if (symbol? f)
               (begin
                  (set! s (assq f data))
                   (if s (cdr s) f)
               )
               f
             )
           )
          )
     )
))

#|
(define (macro-expand-rec  f env)
  (let* ((rec #t) (c 0))
;;;   (format #t "macro-expand-rec ~s  env: ~s~%" f env)
    (set! *alias-name* ())
    (set! *macro-bind-local* ())
   (while rec
      (set! f (syntax-expand f env))
      (set! c (+ c 1))
  ;;;(format #t "syntax-expand  result ~s~%" f )
      (when (= *exflag*  0)
          (set! rec #f)
      )
      (if (> c 500) (syntax-error (cons f c)))
   )
 ;;; (format #t "macro-expand-rec  result ~s~%alias ~s~%" f *alias-name*)
  ;;; (read-char)
  ;;;  2004.3.17
 ;;; (let-define->rec f)   ;;;;;   let define 処理
   ;;;;   recover  (quote .....)
   ;;;;   ex.    (if (eq? x 'if)
   (if (pair? *alias-name*)
       (set! f (quote-recover f *alias-name*))
   )
f
))
|#

;;; 04.9.5
(define (macro-expand-1  f env)
  (let* ((c 0))
  ;;; (format #t "macro-expand-rec ~s  env: ~s~%" f env)
    (set! *alias-name* ())
    ;;;(set! *macro-bind-local* ())
      (set! f (syntax-expand f env))
   ;;;(format #t "macro-expand-rec  result ~s~%" f)
  ;;; (read-char)
  ;;;  2004.3.17
 ;;; (let-define->rec f)   ;;;;;   let define 処
   ;;;;   recover  (quote .....)
   ;;;;   ex.    (if (eq? x 'if)
   (if (pair? *alias-name*)
       (set! f (quote-recover f *alias-name*))
   )
  f
))



(define (s f)
    ( macro-expand-rec f  ())
)


  ;;;;  don't work??? in VER4 !!
;;;;  from spe.ss
#|  toriaezu  2010.7.9
(define-syntax special-syntax (syntax-rules
        (syntax-rules *literal* *macro-key* *macro-bind* )
  ((special-syntax key  (syntax-rules  lite
     (pat1 temp1) (pat2 temp2) ...) )
    ;;;;(let* ((alias (gensym)))
    (let* ((alias (special-alias 'key)))
       (remove-macro 'key)
       (set! *literal*  'lite)
       (set! *macro-key*  'key)
       (set! *macro-bind* (cons (list 'key alias 1 'lite
               (rule-trans key '((pat1 temp1 )(pat2 temp2) ...) ())
               '((pat1 temp1 )(pat2 temp2) ...) ())
                *macro-bind*) )
       (if (symbol-bind? 'key)
            (symbol-value-set  alias (symbol-value 'key))
       )
       (macro-alias 'key)
     )
  ) )
)
|#

;;; overwite macro-expand     -> Seg error!!! why?
;;(shadowing-intern "macro-expand"  "alias")
;;(define (macro-expand form . env)
;;    (if (pair? env) (set! env (car env)))
;;    (syntax-expand form env)
;;)
  ;;;  to be efficient for complete version
#|
(export '(define-syntax  get-macro set-macro remove-macro))
(export '( syntax-rules macro-update macro-symbols macro? ))
(export 'macro-expand-1)
(use-package "macro" "ss")
|#

(set! macro-expand-rec syntax-expand)
#|
(define define-syntax
   (->syntax  (lambda (x)
      (let* ((dummy (car x)) (para (cdr x)) (key (car para)) (exp (cdr para)))
       `(define ,key
          (->syntax ,(car exp) ))
      )
     )
   )
)

(define-syntax define-macro
   (lambda (x)
     `(define-syntax  ,(car (cadr x))
          (let* ((f (lambda ,(cdr (cadr x)) ,@(cddr x))))
            (lambda (exp)
               (apply  f (cdr exp))
             )
          )
   )
))
|#

