package require Tkogl

proc sphere_geometry {} {
    global s_radius s_slices s_stacks
    catch "destroy [winfo children .geom]"
    scale .geom.radius -label radius -from 0.1 -to 3 -res 0.1 \
        -variable s_radius -command redraw -orient horizontal
    scale .geom.slices -label slices -from 3 -to 50 \
        -variable s_slices -command redraw -orient horizontal
    scale .geom.stacks -label stacks -from 3 -to 50 \
        -variable s_stacks -command redraw -orient horizontal
    foreach w [winfo children .geom] {
        pack $w -side top -anchor w
    }
}

proc cylinder_geometry {} {
    global c_baseRadius c_topRadius c_height c_slices c_stacks
    catch "destroy [winfo children .geom]"
    scale .geom.baseradius -label "base radius" -from 0.1 \
        -to 3 -res 0.1 \
        -variable c_baseRadius -command redraw -orient horizontal
    scale .geom.topradius -label "top radius" -from 0.1 \
        -to 3 -res 0.1 \
        -variable c_topRadius -command redraw -orient horizontal
    scale .geom.height -label "height" -from 0.1 \
        -to 3 -res 0.1 \
        -variable c_height -command redraw -orient horizontal
    scale .geom.slices -label slices -from 3 -to 50 \
        -variable c_slices -command redraw -orient horizontal
    scale .geom.stacks -label stacks -from 3 -to 50 \
        -variable c_stacks -command redraw -orient horizontal
    foreach w [winfo children .geom] {
        pack $w -side top -anchor w
    }
}

proc disk_geometry {} {
    global d_innerRadius d_outerRadius d_slices d_loops
    catch "destroy [winfo children .geom]"
    scale .geom.innerradius -label "inner radius" -from 0.0 \
        -to 3 -res 0.1 \
        -variable d_innerRadius -command redraw -orient horizontal
    scale .geom.outerradius -label "outer radius" -from 0.0 \
        -to 3 -res 0.1 \
        -variable d_outerRadius -command redraw -orient horizontal
    scale .geom.slices -label slices -from 3 -to 50 \
        -variable d_slices -command redraw -orient horizontal
    scale .geom.loops -label loops -from 3 -to 50 \
        -variable d_loops -command redraw -orient horizontal
    eval pack [winfo children .geom] -side top -anchor w
}

proc partialdisk_geometry {} {
    global p_innerRadius p_outerRadius p_slices p_loops p_stAngle p_swAngle
    catch "destroy [winfo children .geom]"
    scale .geom.innerradius -label "inner radius" -from 0.0 \
        -to 3 -res 0.1 \
        -variable p_innerRadius -command redraw -orient horizontal
    scale .geom.outerradius -label "outer radius" -from 0.0 \
        -to 3 -res 0.1 \
        -variable p_outerRadius -command redraw -orient horizontal
    scale .geom.slices -label slices -from 3 -to 50 \
        -variable p_slices -command redraw -orient horizontal
    scale .geom.loops -label loops -from 3 -to 50 \
        -variable p_loops -command redraw -orient horizontal
    scale .geom.stangle -label "start angle" -from 0 -to 360 \
        -variable p_stAngle -command redraw -orient horizontal
    scale .geom.swangle -label "sweep angle" -from 0 -to 360 \
        -variable p_swAngle -command redraw -orient horizontal
    eval pack [winfo children .geom] -side top -anchor w
}

proc rotation_bindings { path } {
    global xrot yrot
    set xrot 0 
    set yrot 0
    bind $path <Button-1> { set x %x ; set y %y }
    bind $path <B1-Motion> [subst -nocommands { 
	incr xrot [expr \$x - %x] 
	incr yrot [expr %y - \$y] 
	set x %x
	set y %y
	$path eval -matrixmode modelview \
            -popmatrix \
	    -pushmatrix \
            -rotate \$yrot 1 0 0\
            -rotate \$xrot 0 1 0
	$path redraw
    }]
}

proc quadric_menu { path } {
    global qmenu quad
    set qmenu [tk_optionMenu $path quad sphere \
        cylinder disk partialdisk]
    foreach q {sphere cylinder disk partialdisk} {
	$qmenu entryconfigure $q \
        -command "set quad $q; ${q}_geometry; redraw"
    }
    return $path
}

proc drawstyle_menu { path } {
    global dsmenu drawstyle
    set dsmenu [tk_optionMenu $path drawstyle point line fill silhouette]
    foreach ds {point line fill silhouette} {
	$dsmenu entryconfigure $ds -command "set drawstyle $ds; redraw"
    }
    return $path
}

proc orientation_menu { path } {
    global ormenu orientation
    set ormenu [tk_optionMenu $path orientation inside outside]
    foreach or {inside outside} {
	$ormenu entryconfigure $or -command "set orientation $or; redraw"
    }
    return $path
}

proc normals_menu { path } {
    global nrmenu normals
    set nrmenu [tk_optionMenu $path normals None flat smooth] 
    $nrmenu entryconfigure None -command {
	.gl eval -disable lighting 
	set normals none; redraw
    }
    $nrmenu entryconfigure flat -command {
	.gl eval -enable lighting -shademodel flat
	puts "normals flat"
        set normals flat; redraw
    }
    $nrmenu entryconfigure smooth -command {
	.gl eval -enable lighting -shademodel smooth
	set normals smooth; redraw
    }
    return $path
}

proc texture_menu { path } {
    global txmenu texture
    set txmenu [tk_optionMenu $path texture yes no] 
    $txmenu entryconfigure no -command {
	.gl eval -disable texture2d 
	set texture no; redraw
    }
    $txmenu entryconfigure yes -command {
	.gl eval -enable texture2d 
	set texture yes; redraw
    }
    return $path
}

proc labeled_frame { path labeltxt } {
    frame $path -relief groove -borderwidth 2 
    pack [label $path.label -text $labeltxt] -fill x
    return $path
}

proc createcheckerimg {} {
    image create photo tmp -width 2 -height 2
    image create photo img -width 64 -height 64
    tmp put {{#ffffff #ff0000} {#000000 #ffffff}} 
    img copy tmp -from 0 0 2 2 -zoom 4 4 -to 0 0 64 64
    rename tmp {}
}

proc redraw args {
    global quad texture orientation drawstyle normals
    global qlist
    set parms "-displaylist $qlist -drawstyle \$drawstyle \
        -normals \$normals -orientation \$orientation \
        -texture \$texture"
    switch $quad {
        sphere {
	    global s_radius s_slices s_stacks
            set geom "$s_radius $s_slices $s_stacks"
        }
        cylinder {
	    global c_baseRadius c_topRadius c_height c_slices c_stacks
            set geom "$c_baseRadius $c_topRadius $c_height\
            $c_slices $c_stacks"
        }
        disk { 
	    global d_innerRadius d_outerRadius d_slices d_loops
	    set geom "$d_innerRadius $d_outerRadius $d_slices $d_loops" }
        partialdisk { 
	    global p_innerRadius p_outerRadius p_slices p_loops
	    global p_stAngle p_swAngle
	    set geom "$p_innerRadius $p_outerRadius $p_slices $p_loops\
                      $p_stAngle $p_swAngle" 
	}
    }
    eval .gl $quad $parms $geom
    .gl redraw
}

createcheckerimg

OGLwin .gl -width 400 -height 400

.gl eval \
	-matrixmode projection \
	-loadidentity \
	-ortho -5 5 -5 5 -5 5 \
        -light light0 position 10 10 10 \
	-enable lighting \
	-enable light0 \
	-enable colormaterial \
	-enable depthtest \
	-texparameter texture2d texturewraps repeat \
	-texparameter texture2d texturewrapt repeat \
	-texparameter texture2d texturemagfilter nearest \
	-texparameter texture2d textureminfilter nearest \
	-texenv textureenv textureenvmode modulate\
        -teximage2d 0 0 img

set qlist [.gl newlist]
set quad sphere
set drawstyle fill
set normals smooth
set texture no
set orientation outside
set s_radius 1
set s_stacks 10
set s_slices 10
set c_baseRadius 1.0
set c_topRadius 1.0
set c_height 1.0
set c_stacks 10
set c_slices 10
set d_innerRadius 0
set d_outerRadius 1
set d_slices 10
set d_loops 3
set p_innerRadius 0
set p_outerRadius 1
set p_slices 10
set p_loops 3
set p_stAngle 0
set p_swAngle 270

pack [frame .cmd] -side bottom -fill x

pack .gl -expand yes -side left -fill both

pack [button .cmd.quit -text quit -command exit] -fill y -side left

pack [labeled_frame .cmd.quad Quadric] -side left
pack [quadric_menu .cmd.quad.menu]

pack [labeled_frame .cmd.ds "Display Style"] -side left
pack [drawstyle_menu .cmd.ds.menu] 

pack [labeled_frame .cmd.orient "Normal Orientation"] -side left
pack [orientation_menu .cmd.orient.menu]

pack [labeled_frame .cmd.normals "Normal Generation"] -side left
pack [normals_menu .cmd.normals.menu]

pack [labeled_frame .cmd.texture "Texture"] -side left
pack [texture_menu .cmd.texture.menu]

pack [frame .geom -borderwidth 3 -relief sunken] -side left -fill x

.gl main -clear colorbuffer depthbuffer\
         -call $qlist

rotation_bindings .gl

$qmenu invoke $quad 
